
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <windows.h>
#include <ras.h>
#include <exception>
#include "ntp-service.h"
#include "wrap-winsock.h"
#include "enable-set-time.h"
#include "raise-priority.h"
#include "ntp-packet.h"
#include "logging.h"
#include "wrap-socket.h"
#include "regstuff.h"
#include "lasterr-exception.h"

typedef DWORD RasEnumConnections_func(LPRASCONN,LPDWORD,LPDWORD);
typedef DWORD RasGetConnectStatus_func(HRASCONN,LPRASCONNSTATUS);

char * NtpService::ntpversion = "2.3";

NtpService::NtpService(HINSTANCE init_hInstance)
: WrapService(init_hInstance,
              "NTPTime",
              SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN),
  logger(NULL), ras_dll_loaded(false), ras_dll_handle(NULL)
{
   stop_event = CreateEvent(NULL, FALSE, FALSE, NULL);
}

NtpService::~NtpService()
{
   CloseHandle(stop_event);
   delete logger;
}

void
NtpService::Run()
{
   WrapCriticalSection wcs(run_lock);
   try {
      while (! shutting_down) {
         if (wait_for_ras) {
            while (! IsRasActive()) {
               if (WaitForSingleObject(stop_event, ras_poll_milliseconds) !=
                   WAIT_TIMEOUT) {
                  shutting_down = true;
                  break;
               }
            }
         }
         if (shutting_down) {
            break;
         }
         TryOnce();
         if (! shutting_down) {
            if (WaitForSingleObject(stop_event, ntp_poll_milliseconds) !=
                WAIT_TIMEOUT) {
               shutting_down = true;
               break;
            }
         } else {
            break;
         }
      }
   } catch (exception & ee) {
      shutting_down = true;
      LogStuff * log = GetLog();
      log->LogException(ee);
      log->LogString("shutting down NTP service");
   }
}

bool
NtpService::OnInit()
{
   bool rval = true;
   log_directory_name = NULL;
   write_log = false;
   ntp_server_port = 123; // default NTP port number
   max_adjust_seconds = 0; // default no limit
   try {
      shutting_down = false;
      RegistryLocalMachine lm;
      RegistryKey ntproot(lm, "Software\\ntcli\\NTPTime");
      RegistryKey ntpkey(ntproot, ntpversion);
      DWORD temp;
      temp = ntpkey.GetRegDwordValue("WriteLogEntries");
      if (temp != 0) {
         log_directory_name = ntpkey.GetRegSzValue("LogDirectory");
      }
      write_log = (temp != 0);
      temp = ntpkey.GetRegDwordValue("WaitForRAS");
      wait_for_ras = (temp != 0);
      if (wait_for_ras) {
         ras_poll_milliseconds =
            ntpkey.GetRegDwordValue("RASPollMilliseconds");
      }
      temp = ntpkey.GetRegDwordValue("RunOnce");
      run_once = (temp != 0);
      ntp_server_name = ntpkey.GetRegSzValue("NTPServerName");
      ntp_poll_milliseconds = ntpkey.GetRegDwordValue("NTPPollMilliseconds");
      DWORD fmt;
      DWORD len;
      if ((ntpkey.QueryValue("NTPServerPort", fmt, len)) &&
          (fmt == REG_DWORD)) {
         ntp_server_port = (unsigned short)
            ntpkey.GetRegDwordValue("NTPServerPort");
      }
      if ((ntpkey.QueryValue("MaxAdjustSeconds", fmt, len)) &&
          (fmt == REG_DWORD)) {
         max_adjust_seconds = ntpkey.GetRegDwordValue("MaxAdjustSeconds");
      }
      GetLog()->LogPrintf("Starting up NTP service version %s, server %s",
                             ntpversion, ntp_server_name);
   } catch (exception & ee) {
      rval = false;
      GetLog()->LogException(ee);
   }
   return rval;
}

void
NtpService::OnStop()
{
   shutting_down = true;
   SetEvent(stop_event);
   {
      WrapCriticalSection wcs(run_lock);
   }
   GetLog()->LogString("NTP stop request");
}

void
NtpService::OnInterrogate()
{
}

void
NtpService::OnPause()
{
}

void
NtpService::OnContinue()
{
}

void
NtpService::OnShutdown()
{
   shutting_down = true;
   SetEvent(stop_event);
   {
      WrapCriticalSection wcs(run_lock);
   }
   GetLog()->LogString("NTP shutdown request");
}

bool
NtpService::OnUserControl(DWORD user_opcode)
{
   return true;
}

bool
NtpService::IsRasActive()
{
   // Look at all the trouble we have to go to to find out if there is an
   // active RAS connection while avoiding autodial (There is actually a
   // wonderful service that will allow me to wait on an event, but it only
   // works under NT 4.0 or greater, sigh...)

   // Now we go to even *more* trouble to avoid having to have RAS installed
   // in case this routine is never called (A lot of folks have direct
   // network connections and don't want to go to the trouble of installing
   // RAS on their workstations). The first time we call this, try to
   // LoadLibrary the rasapi32 lib and dynamically lookup the RAS functions
   // we need to call...

   if (! ras_dll_loaded) {
      ras_dll_loaded = true;
      ras_dll_handle = LoadLibrary("rasapi32");
      if (ras_dll_handle == NULL) {
         logger->LogString("Unable to LoadLibrary the rasapi32 DLL");
         return false;
      }
      ras_enum_connections_ptr =
         GetProcAddress(ras_dll_handle, "RasEnumConnectionsA");
      ras_get_connection_status_ptr =
         GetProcAddress(ras_dll_handle, "RasGetConnectStatusA");
   }
   if ((ras_dll_handle == NULL) || (ras_enum_connections_ptr == NULL) ||
       (ras_get_connection_status_ptr == NULL)) {
      return false;
   }
   RasEnumConnections_func * RasEnumConnections =
      (RasEnumConnections_func *)ras_enum_connections_ptr;
   RasGetConnectStatus_func * RasGetConnectStatus =
      (RasGetConnectStatus_func *)ras_get_connection_status_ptr;
   RASCONN * connection_poop;
   RASCONN dummy;
   DWORD connection_count;
   DWORD connection_size = 0;
   memset((void *)&dummy, 0, sizeof(dummy));
   dummy.dwSize = sizeof(RASCONN);
   RasEnumConnections(&dummy, &connection_size, &connection_count);
   if (connection_size < sizeof(RASCONN)) {
      connection_size = sizeof(RASCONN);
   }
   connection_count = (connection_size +
                       sizeof(connection_poop[0]) - 1) /
                          sizeof(connection_poop[0]);
   connection_poop = new RASCONN [connection_count];
   connection_size = connection_count * sizeof(connection_poop[0]);
   memset((void *)connection_poop, 0, connection_size);
   connection_poop[0].dwSize = sizeof(RASCONN);
   if (RasEnumConnections(connection_poop, &connection_size,
                          &connection_count) != 0) {
      delete [] connection_poop;
      LastErrorException ee("RasEnumConnections");
      throw ee;
   }
   bool found_valid_ras = false;
   for (DWORD ri = 0; ri < connection_count; ++ri) {
      RASCONNSTATUS ras_stat;
      memset((void *)&ras_stat, 0, sizeof(ras_stat));
      ras_stat.dwSize = sizeof(ras_stat);
      if (RasGetConnectStatus(connection_poop[ri].hrasconn, &ras_stat) == 0) {
         if (ras_stat.rasconnstate == RASCS_Connected) {
            found_valid_ras = true;
            break;
         }
      }
   }
   delete [] connection_poop;
   return found_valid_ras;
}

void
NtpService::TryOnce()
{
   SYSTEMTIME right_now;
   WrapWinsock ww;   // Initialize winsock services...

   struct sockaddr_in serv_addr;
   memset((void *)&serv_addr, 0, sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_port = htons(ntp_server_port);
   struct hostent * hp = gethostbyname(ntp_server_name);
   if (hp == NULL) {
      LastErrorException ee("gethostbyname",
                            "Looking up NTP server name, %a: %r");

      // Don't barf just yet. Maybe this is an absolute IP address string...

      unsigned long ipaddr = inet_addr(ntp_server_name);
      if (ipaddr != INADDR_NONE) {
         memcpy((void *)&serv_addr.sin_addr.s_addr, (void *)&ipaddr,
                sizeof(serv_addr.sin_addr.s_addr));
      } else {
         throw ee;
      }
   } else {
      memcpy((void *)&serv_addr.sin_addr.s_addr, (void *)hp->h_addr,
             sizeof(serv_addr.sin_addr.s_addr));
   }

   WrapSocket s(socket(AF_INET, SOCK_DGRAM, 0));
   if ((SOCKET)s == INVALID_SOCKET) {
      LastErrorException ee("socket");
      throw ee;
   }

   struct sockaddr_in addr;
   memset((char *)&addr, 0, sizeof addr);
   addr.sin_family = AF_INET;
   addr.sin_port = htons(ntp_server_port);
   addr.sin_addr.s_addr = htonl(INADDR_ANY);
   if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR) {
      LastErrorException ee("bind");
      throw ee;
   }

   int on = 1;
   if (ioctlsocket(s, FIONBIO, (u_long *) &on) == SOCKET_ERROR) {
      LastErrorException ee("ioctlsocket");
      throw ee;
   }

   struct timeval ten;
   ten.tv_sec = 10;
   ten.tv_usec = 0;
   fd_set reader;
   FD_ZERO(&reader);
   FD_SET(s, &reader);

   RaisePriority rp; // raise process priority to max while diddling time

   BasicNtpPacket cpkt;
   FullNtpPacket  spkt;
   memset((void *)&cpkt, 0, sizeof(cpkt));
   SetPacketLeapVersionMode(cpkt,0,3,3);
   GetSystemTime(&right_now);
   NtpLongFixed rn(right_now);
   rn.ConvertToNetworkOrder(cpkt.transmit_timestamp);

   if (sendto(s, (char *)&cpkt, sizeof(cpkt), 0,
        (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == SOCKET_ERROR) {
      LastErrorException ee("sendto");
      throw ee;
   }

   int select_status = select(1, &reader, NULL, NULL, &ten);
   if (select_status == SOCKET_ERROR) {
      LastErrorException ee("select");
      throw ee;
   }
   SYSTEMTIME recvtime;
   GetSystemTime(&recvtime);
   if (select_status == 0) {
      // I seem to have timed out without getting a return packet. Just
      // log it and get outta here
      GetLog()->LogString("no response from NTP server");
      return;
   }
   struct sockaddr_in gotfrom;
   int gotlen = sizeof(gotfrom);
   if (recvfrom(s, (char *)&spkt, sizeof(spkt), 0,
          (struct sockaddr *)&gotfrom, &gotlen) == SOCKET_ERROR) {
      LastErrorException ee("recvfrom");
      throw ee;
   }

   // PUT CODE HERE - need to do minimal validation of incoming NTP packet.

   NtpLongFixed T1(spkt.basic_part.originate_timestamp);
   NtpLongFixed T2(spkt.basic_part.receive_timestamp);
   NtpLongFixed T3(spkt.basic_part.transmit_timestamp);
   NtpLongFixed T4(recvtime);

   NtpLongFixed d = (T4 - T1) - (T2 - T3);
   NtpLongFixed t = ((T2 - T1) + (T3 - T4)).SignedDivByTwo();

   if (max_adjust_seconds > 0) {
      // If there is a configured max adjustment limit, check it
      NtpLongFixed m(t);
      if (m.IntegerPart() & 0x80000000) {
         m = - m;
      }
      if (m.IntegerPart() > max_adjust_seconds) {
         GetLog()->LogPrintf("time adjustment of %ld seconds > max of %ld",
                                m.IntegerPart(), max_adjust_seconds);
         return;
      }
   }

   EnableSetTime est; // Diddle privileges so I can set the time...

   GetSystemTime(&right_now);
   NtpLongFixed new_clock(right_now);
   new_clock = new_clock + t;
   new_clock.ConvertToSYSTEMTIME(right_now);
   if (! SetSystemTime(&right_now)) {
      LastErrorException ee("SetSystemTime");
      throw ee;
   }
   LogTimeDiff("clock adjustment", t);
   LogTimeDiff("   round trip delay", d);

   // I seem to have actually set the time. If I was only supposed to do this
   // once, then trigger a shutdown.

   if (run_once) {
      shutting_down = true;
   }

   // At this return (or at any 'throw' above), all the destructors should
   // tun back off all the privileges, raised priority, activated winsock,
   // etc.

   return;
}

LogStuff *
NtpService::GetLog()
{
   if (logger == NULL) {
      if (write_log) {
         logger = new LogDirectory(log_directory_name, "ntp");
      } else {
         logger = new LogStuff;
      }
   }
   return logger;
}

LPCTSTR
NtpService::GetDependencyList()
{
   // If RAS is installed on this system, then make sure we depend on it so
   // it will get started before we start trying to poll the RAS
   // configuration info. If RAS isn't installed, I guess we'll be using a
   // direct internet connection, so we won't try to call the RAS polling
   // code anyway...

   if (IsInstalled("RasMan")) {
      return "RasMan\0\0";
   } else {
      return WrapService::GetDependencyList();
   }
}

void
NtpService::LogTimeDiff(const char * msg, NtpLongFixed & t)
{
   int neg = 0;
   NtpLongFixed x(t);
   if (t.IntegerPart() & 0x80000000) {
      neg = 1;
      x = - x;
   }
   SYSTEMTIME st;
   NtpLongFixed fx(0, x.FractionalPart());
   fx.ConvertToSYSTEMTIME(st);
   GetLog()->LogPrintf("%s: %s%ld.%03d seconds",
                          msg, (neg?"-":""), x.IntegerPart(),
                             st.wMilliseconds);
}
