
// Portions Copyright (C) Tom Horsley, 1997. All rights reserved.
// Other portions adapted from the C++ server class sample code from the
// MSDN library.

#include "wrap-service.h"
#include "wrap-sc-handle.h"
#include "logging.h"

WrapService* WrapService::this_service = NULL;

WrapService::WrapService(
   HINSTANCE init_hInstance,
   const char* init_service_name,
   DWORD init_controls_accepted)
: service_name(init_service_name),
  service_status(NULL),
  is_running_on_nt(false),
#ifdef ENABLE_DEBUG
  run_on_console(false),
#endif
  already_stopped(false),
  dummy_log(NULL),
  hInstance(init_hInstance)
{
   OSVERSIONINFO osvi;
   osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
   if (GetVersionEx(&osvi)) {
      is_running_on_nt = (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT);
   }

   this_service = this;

   status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
   status.dwCurrentState = SERVICE_STOPPED;
   status.dwControlsAccepted = init_controls_accepted;
   status.dwWin32ExitCode = 0;
   status.dwServiceSpecificExitCode = 0;
   status.dwCheckPoint = 0;
   status.dwWaitHint = 0;
}

WrapService::~WrapService()
{
   return;
}

bool
WrapService::IsInstalled(const char * alt_name)
{
   if (! IsRunningOnNT()) {
      // On 95, if you can run the program, it must be installed :-).
      return true;
   }
   if (alt_name == NULL) {
      alt_name = service_name;
   }
   bool rval = false;
   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm != NULL) {
      WrapScHandle service(OpenService(scm, alt_name,
                                          SERVICE_QUERY_CONFIG));
      if (service != NULL) {
         rval = true;
      }
   }
   return rval;
}

bool
WrapService::Install(DWORD start_type)
{
   if (! IsRunningOnNT()) {
      // On 95, nothing magical needed to install.
      return true;
   }
   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm == NULL) return false;

   char file_path[_MAX_PATH];
   GetModuleFileName(NULL, file_path, sizeof(file_path));

   WrapScHandle service(CreateService(scm,
                                      service_name,
                                      service_name,
                                      SERVICE_ALL_ACCESS,
                                      SERVICE_WIN32_OWN_PROCESS,
                                      start_type,
                                      SERVICE_ERROR_NORMAL,
                                      file_path,
                                      NULL,
                                      NULL,
                                      GetDependencyList(),
                                      NULL,
                                      NULL));
   if (service == NULL) {
      return false;
   }
   return true;
}

LPCTSTR
WrapService::GetDependencyList()
{
   return NULL;
}

bool
WrapService::Uninstall()
{
   if (! IsRunningOnNT()) {
      // On 95, nothing magical needed to un-install...
      return true;
   }
   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm == NULL) return false;

   bool rval = false;
   WrapScHandle service(OpenService(scm, service_name, DELETE));
   if (service != NULL) {
      if (DeleteService(service)) {
         rval = true;
      }
   }
   return rval;
}

bool
WrapService::StartService()
{
   SERVICE_TABLE_ENTRY st[] = {
      {(char *)service_name, ServiceMain},
      {NULL, NULL}
   };

   if (IsRunningOnNT()) {
      // On NT, go through normal service dispatcher rigamarole.
      return (StartServiceCtrlDispatcher(st) != 0);
   } else {

      // On 95, start the "service" in another thread and run a window loop
      // on a hidden window in the main thread to listen for close and
      // shutdown messages.

      DWORD thread_id;
      CreateThread(NULL, 0, Win95ServiceThread, NULL, 0, &thread_id);

      WNDCLASSEX  wndclass;
      memset((void *)&wndclass, 0, sizeof(wndclass));
      wndclass.cbSize = sizeof(wndclass);
      wndclass.style = CS_HREDRAW | CS_VREDRAW;
      wndclass.lpfnWndProc = Win95WindowProc;
      wndclass.hInstance = hInstance;
      wndclass.lpszClassName = service_name;
      RegisterClassEx(&wndclass);
      HWND hwnd;
      hwnd = CreateWindowEx(WS_EX_TOOLWINDOW,
                            service_name,
                            service_name,
                            WS_POPUP | WS_SIZEBOX | WS_SYSMENU,
                            0, 0, 0, 0,
                            NULL,
                            NULL,
                            hInstance,
                            NULL);
      ShowWindow(hwnd, SW_HIDE);
      UpdateWindow(hwnd);
      MSG msg;
      while (GetMessage(&msg, NULL, 0, 0)) {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
      return true;
   }
   return false;
}

void
WrapService::SetStatus(DWORD new_status)
{
   WrapCriticalSection wcs(stat_lock);
   if (! ((new_status == SERVICE_STOPPED) && already_stopped)) {
      status.dwCurrentState = new_status;
      if (IsRunningOnNT()) {
         if (status.dwCurrentState == SERVICE_STOPPED) {
            already_stopped = true;
            if (status.dwServiceSpecificExitCode == 0) {
               status.dwServiceSpecificExitCode = 1;
            }
         }
         SetServiceStatus(service_status, &status);
      }
   }
}

bool
WrapService::Initialize()
{
   SetStatus(SERVICE_START_PENDING);
   
   bool rval = OnInit(); 
   
   status.dwWin32ExitCode = GetLastError();
   status.dwCheckPoint = 0;
   status.dwWaitHint = 0;
   if (! rval) {
      SetStatus(SERVICE_STOPPED);
      return false;   
   }
   
   SetStatus(SERVICE_RUNNING);

   return true;
}

bool
WrapService::OnInit()
{
   return true;
}

void
WrapService::OnStop()
{
   return;
}

void
WrapService::OnInterrogate()
{
   return;
}

void
WrapService::OnPause()
{
   return;
}

void
WrapService::OnContinue()
{
   return;
}

void
WrapService::OnShutdown()
{
   return;
}

bool
WrapService::OnUserControl(DWORD /* user_opcode */)
{
   return false;
}

LogStuff *
WrapService::GetLog()
{
   if (dummy_log == NULL) {
      dummy_log = new LogStuff;
   }
   return dummy_log;
}

#ifdef ENABLE_DEBUG
bool
WrapService::Debugging(int /* level */)
{
   return false;
}

void
WrapService::RunOnConsole()
{
   run_on_console = true;
}
#endif

void WINAPI
WrapService::ServiceMain(DWORD /* argc */, LPTSTR* /* argv */)
{
   this_service->status.dwCurrentState = SERVICE_START_PENDING;
   this_service->service_status =
      RegisterServiceCtrlHandler(this_service->service_name, Handler);

   if (this_service->service_status == NULL) {
      return;
   }

   if (this_service->Initialize()) {
      this_service->status.dwWin32ExitCode = 0;
      this_service->status.dwCheckPoint = 0;
      this_service->status.dwWaitHint = 0;
      this_service->Run();
   }

   this_service->SetStatus(SERVICE_STOPPED);
}

void WINAPI
WrapService::Handler(DWORD opcode)
{
   switch (opcode) {
   case SERVICE_CONTROL_STOP:
      this_service->SetStatus(SERVICE_STOP_PENDING);
      this_service->OnStop();
      break;

   case SERVICE_CONTROL_PAUSE:
      this_service->OnPause();
      break;

   case SERVICE_CONTROL_CONTINUE:
      this_service->OnContinue();
      break;

   case SERVICE_CONTROL_INTERROGATE:
      this_service->OnInterrogate();
      break;

   case SERVICE_CONTROL_SHUTDOWN:
      this_service->OnShutdown();
      break;

   default:
      this_service->OnUserControl(opcode);
      break;
   }
   this_service->SetStatus(this_service->status.dwCurrentState);
}

bool WrapService::w95_shutdown = false;

DWORD WINAPI
WrapService::Win95ServiceThread(LPVOID /* arg */)
{
   if (this_service->OnInit()) {
      this_service->Run();
      this_service->SetStatus(SERVICE_STOPPED);
   }
   return 0;
}

LRESULT CALLBACK
WrapService::Win95WindowProc(
   HWND hwnd,
   UINT iMsg,
   WPARAM wParam,
   LPARAM lParam)
{
   switch(iMsg) {
   case WM_DESTROY:
      this_service->OnShutdown();
      PostQuitMessage(0);
      return 0;
   }
   return DefWindowProc(hwnd, iMsg, wParam, lParam);
}
