#ifndef wrap_service_h
#define wrap_service_h

// Portions Copyright (C) Tom Horsley, 1997. All rights reserved.
// Other portions adapted from the C++ server class sample code from the
// MSDN library.

#include <windows.h>
#include "wrap-critical.h"

class LogStuff;

// An NT "service" wrapper class adapted from sample code in VC++. Only 1 of
// these objects may be created in any given program.
//
// On Windows 95, most of the routines in here do nothing, but the
// StartService routine will invoke OnInit() followed by Run() so that code
// using this class can do the same work as a service when running as an
// ordinary program under Windows 95.
//
class WrapService {
public:

       // Constructor wants the name of the service.
       //
   WrapService(
      HINSTANCE init_hInstance,
      const char* init_service_name,
      DWORD init_controls_accepted = SERVICE_ACCEPT_STOP);

      // Destructor frees up resources.
      //
   virtual
   ~WrapService();

      // TRUE if running on windows NT where an actual service can be
      // implemented, FALSE if running on Windows 95 where most functions do
      // nothing. If you plan to write a Run() function that needs to
      // interact with the service manager, then you'll want to check this
      // dynamically.
      //
   bool
   IsRunningOnNT();

      // Check to see if this service is installed on the system already.
      // Normally asks about this service, but you can pass an alternate
      // service name to check on (possibly useful for dynamically deciding
      // on dependencies).
      //
   bool
   IsInstalled(const char * alt_name = NULL);

      // Install the current program as a service.
      //
   bool
   Install(DWORD start_type = SERVICE_DEMAND_START);

      // If the service which is to be installed depends on other services,
      // then this function should be defined in the subclass to return a
      // string containing NULL separated names of the dependencies (with a
      // double null byte termination). The default is just NULL.
      //
   virtual LPCTSTR
   GetDependencyList();

      // Remove the service.
      //
   bool
   Uninstall();

      // Start this sucker up (this should be called by the main program of
      // the service to actually start running as a service - prior to this
      // call, you're just an ordinary program).
      //
   bool
   StartService();

      // Communicate the current service status to the service manager
      //
   void
   SetStatus(DWORD new_status);

      // Initialize the service. (This is called by the "ServiceMain"
      // function to do standard initialization work. It calls OnInit()
      // which is the function derived classes should override).
      //
   bool
   Initialize();

      // This function is called to do the real work when the service is
      // running (You MUST override this to implement a specfic kind of
      // service). The way NT services are implemented, this function will
      // run in a separate thread from the OnWhatever() functions that
      // process service control requests.
      //
   virtual void
   Run() = 0;

      // This function is called to do any initialization required by
      // the derived class when the service is initialized.
      //
   virtual bool
   OnInit();

      // Override this to do any required actions when the service is
      // stopped.
      //
   virtual void
   OnStop();

      // Override this to do any required actions when the service is
      // interrogated.
      //
   virtual void
   OnInterrogate();

      // Override this to do any required actions when the service is
      // paused.
      //
   virtual void
   OnPause();

      // Override this to do any required actions when the service is
      // continued.
      //
   virtual void
   OnContinue();

      // Override this to do any required actions when the service is
      // shutdown.
      //
   virtual void
   OnShutdown();

      // Override this to do any required actions when the service is
      // passed a user control command.
      //
   virtual bool
   OnUserControl(DWORD user_opcode);

      // Call this routine to get a pointer to a class you may use to
      // log errors, debug output, etc.
      //
   virtual LogStuff *
   GetLog();

#ifdef ENABLE_DEBUG
      // Call this to find out if the debug level is 'level' or greater.
      //
   virtual bool
   Debugging(int level);

      // Call this to run as console app (like on 95) even if on NT.
      //
   virtual void
   RunOnConsole();
#endif

      // static member functions - pointers to these are used by the
      // service control manager to control the server.
      //
   static void WINAPI
   ServiceMain(DWORD argc, LPTSTR* argv);

   static void WINAPI
   Handler(DWORD opcode);

protected:

      // data members
      //
   const char * service_name;
   SERVICE_STATUS_HANDLE service_status;
   SERVICE_STATUS status;
   bool is_running_on_nt;
#ifdef ENABLE_DEBUG
   bool run_on_console;
#endif
   DeclareCriticalSection stat_lock;
   bool already_stopped;
   LogStuff * dummy_log;
   HINSTANCE hInstance;

      // static data
      //
   static WrapService* this_service;
   static bool w95_shutdown;

      // Thread to run service in.
      //
   static DWORD WINAPI
   Win95ServiceThread(LPVOID arg);

      // Windows procedure used when running on 95 or 98.
      //
   static LRESULT CALLBACK
   Win95WindowProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam);
};

inline bool
WrapService::IsRunningOnNT()
{
   return is_running_on_nt
#ifdef ENABLE_DEBUG
             && ! run_on_console
#endif
                          ;
}

#endif // wrap_service_h
